library docs_analyzer.bin;

import 'dart:io';

import 'package:docs_analyzer/docs_analyzer.dart';

import 'package:args/args.dart';
import 'package:cli_util/cli_util.dart' as cli_util;

main(List<String> arguments) async {
  var parser = _createArgsParser();
  ArgResults args;
  try {
    args = parser.parse(arguments);
  } on FormatException catch (e) {
    _printHelp(parser, error: e);
  }

  if (args['help']) {
    _printHelp(parser);
  }

  Directory sdkDir = cli_util.getSdkDir(arguments);
  Directory inputDir = new Directory(args['directory']);
  bool listUndocumented = args['list-undocumented'];
  List<String> excludes = args['exclude'];
  bool showSummary = args['summary'];

  var docsAnalysis = await analyzeDocumentation(inputDir, sdkDir, excludes);

  if (showSummary) {
    _printSummaryAndExit(docsAnalysis);
  }

  if (listUndocumented) {
    docsAnalysis.undocumentedElements.forEach((elem) => print(elem.fullyQualifiedName));
  }

  if (docsAnalysis.missingDocumentation && !listUndocumented) {
    print('Some entities are missing documentation. Run with --list-undocumented to list them all');
  }

  int exitCode = docsAnalysis.missingDocumentation ? 1 : 0;
  exit(exitCode);
}

void _printSummaryAndExit(DocsAnalysis docsAnalysis) {
  int numElements = docsAnalysis.allElements.length;
  int numUndocumented = docsAnalysis.undocumentedElements.length;

  if (!docsAnalysis.missingDocumentation) {
    print('Congratulations! All $numElements public entities are documented.');
    exit(0);
  }

  print('${numUndocumented} out of ${numElements} entities are undocumented. '
        'Distribution of undocumented entities:');
  print('    Classes:   ${docsAnalysis.undocumentedClasses.length}');
  print('    Functions: ${docsAnalysis.undocumentedFunctions.length}');
  print('    Variables: ${docsAnalysis.undocumentedVariables.length}');
  print('    Enums:     ${docsAnalysis.undocumentedEnums.length}');

  exit(0);
}

ArgParser _createArgsParser() {
  var parser = new ArgParser();
  parser.addFlag('help', abbr: 'h', negatable: false,
      help: 'Show command help.');
  parser.addFlag('summary', negatable: false,
      help: 'Show documentation statistics.');
  parser.addOption('directory', defaultsTo: Directory.current.path,
      help: 'Path to source directory.');
  parser.addFlag('list-undocumented', defaultsTo: false,
      help: 'List all elements lacking documentation.');
  parser.addOption('exclude', allowMultiple: true,
      help: 'Libraries to exclude from analysis.');
  return parser;
}

void _printHelp(ArgParser parser, {FormatException error: null}) {
  if (error != null) {
    print(error.message);
    print('');
  } else {
    print('Checks dart libraries for missing documentation.\n');
  }

  print('Usage: dartdoc [OPTIONS]\n');
  print(parser.usage);
  exit(error == null ? 0 : 64);
}


