library docs_analyzer;

import 'dart:async';
import 'dart:io';

import 'package:analyzer/file_system/file_system.dart' as fileSystem;
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:analyzer/source/package_map_resolver.dart';
import 'package:analyzer/source/pub_package_map_provider.dart';
import 'package:analyzer/source/sdk_ext.dart';
import 'package:analyzer/src/generated/element.dart';
import 'package:analyzer/src/generated/engine.dart';
import 'package:analyzer/src/generated/error.dart';
import 'package:analyzer/src/generated/java_io.dart';
import 'package:analyzer/src/generated/sdk.dart';
import 'package:analyzer/src/generated/sdk_io.dart';
import 'package:analyzer/src/generated/source.dart';
import 'package:analyzer/src/generated/source_io.dart';

import 'package:dartdoc/src/io_utils.dart';
import 'package:dartdoc/src/model.dart';
import 'package:dartdoc/src/package_meta.dart';

export 'package:dartdoc/src/element_type.dart';
export 'package:dartdoc/src/generator.dart';
export 'package:dartdoc/src/model.dart';
export 'package:dartdoc/src/package_meta.dart';

Future<DocsAnalysis> analyzeDocumentation(
    Directory inputDir, Directory sdkDir, List<String> excludes) async {
  PackageMeta packageMeta = new PackageMeta.fromDir(inputDir);
  DartDoc doc = new DartDoc(inputDir, excludes, sdkDir, packageMeta);
  Package results = await doc.generateDocs();

  return new DocsAnalysis(results);
}

/// Generates Dart documentation for all public Dart libraries in the given
/// directory.
class DartDoc {
  final Directory rootDir;
  final List<String> excludes;
  final Directory sdkDir;
  final PackageMeta packageMeta;

  DartDoc(this.rootDir, this.excludes, this.sdkDir, this.packageMeta);

  /// Generate DartDoc documentation.
  ///
  /// [DartDocResults] is returned if dartdoc succeeds. [DartDocFailure] is thrown if dartdoc fails
  /// in an expected way, for example if there is an analysis error in the code. Any other exception
  /// can be thrown if there is an unexpected failure.
  Future<Package> generateDocs() async {
    List<String> files = findFilesToDocumentInPackage(rootDir.path).toList();

    List<LibraryElement> libraries = _parseLibraries(files);

    // remove excluded libraries
    excludes.forEach((pattern) {
      libraries.removeWhere((lib) {
        return lib.name.startsWith(pattern) || lib.name == pattern;
      });
    });

    return new Package(libraries, packageMeta);
  }

  List<LibraryElement> _parseLibraries(List<String> files) {
    Set<LibraryElement> libraries = new Set();
    DartSdk sdk = new DirectoryBasedDartSdk(new JavaFile(sdkDir.path));
    List<UriResolver> resolvers = [new DartUriResolver(sdk)];

    fileSystem.Resource cwd =
        PhysicalResourceProvider.INSTANCE.getResource(rootDir.path);

    var pubPackageMapProvider =
        new PubPackageMapProvider(PhysicalResourceProvider.INSTANCE, sdk);

    var packageMapInfo = pubPackageMapProvider.computePackageMap(cwd);

    Map<String, List<fileSystem.Folder>> packageMap = packageMapInfo.packageMap;
    if (packageMap != null) {
      resolvers.add(new SdkExtUriResolver(packageMap));
      resolvers.add(new PackageMapUriResolver(
          PhysicalResourceProvider.INSTANCE, packageMap));
    }
    resolvers.add(new FileUriResolver());

    SourceFactory sourceFactory = new SourceFactory(resolvers);

    var options = new AnalysisOptionsImpl()..cacheSize = 512;

    AnalysisContext context = AnalysisEngine.instance.createAnalysisContext()
      ..analysisOptions = options
      ..sourceFactory = sourceFactory;

    List<Source> sources = [];

    files.forEach((String filePath) {
      String name = filePath;
      if (name.startsWith(Directory.current.path)) {
        name = name.substring(Directory.current.path.length);
        if (name.startsWith(Platform.pathSeparator)) name = name.substring(1);
      }

      JavaFile javaFile = new JavaFile(filePath);
      Source source = new FileBasedSource(new JavaFile(filePath));
      Uri uri = context.sourceFactory.restoreUri(source);
      if (uri != null) {
        source = new FileBasedSource(javaFile, uri);
      }
      sources.add(source);
      if (context.computeKindOf(source) == SourceKind.LIBRARY) {
        LibraryElement library = context.computeLibraryElement(source);
        libraries.add(library);
      }
    });

    // Ensure that the analysis engine performs all remaining work.
    AnalysisResult result = context.performAnalysisTask();
    while (result.hasMoreWork) {
      result = context.performAnalysisTask();
    }

    List<AnalysisErrorInfo> errorInfos = [];

    for (Source source in sources) {
      context.computeErrors(source);
      errorInfos.add(context.getErrors(source));
    }

    List<_Error> errors = errorInfos.expand((AnalysisErrorInfo info) {
      return info.errors.map(
          (error) => new _Error(error, info.lineInfo, packageMeta.dir.path));
    }).where((_Error error) => error.isError).toList()..sort();

    if (errors.isNotEmpty) {
      errors.forEach(print);
      int len = errors.length;
      throw new DartDocFailure(
          "encountered ${len} analysis error${len == 1 ? '' : 's'}");
    }

    return libraries.toList();
  }
}

class DocsAnalysis {
  final Package package;

  List<ModelElement> allElements;
  List<ModelElement> undocumentedElements;

  List<Class> classes = [];
  List<Class> enums = [];
  List<TopLevelVariable> variables = [];
  List<ModelFunction> functions = [];

  List<Class> undocumentedClasses = [];
  List<Class> undocumentedEnums = [];
  List<TopLevelVariable> undocumentedVariables = [];
  List<ModelFunction> undocumentedFunctions = [];

  bool get missingDocumentation => undocumentedElements.length != 0;

  DocsAnalysis(this.package) {
    package.libraries.forEach((lib) {
      classes.addAll(lib.allClasses);
      enums.addAll(lib.enums);
      variables.addAll(lib.properties);
      variables.addAll(lib.constants);
      functions.addAll(lib.functions);
    });

    allElements = [classes, enums, variables, functions].expand((e) => e).toList();
    undocumentedElements = allElements.where(_elementLacksDocumentation).toList();

    undocumentedClasses = classes.where(_elementLacksDocumentation).toList();
    undocumentedEnums = enums.where(_elementLacksDocumentation).toList();
    undocumentedVariables = variables.where(_elementLacksDocumentation).toList();
    undocumentedFunctions = functions.where(_elementLacksDocumentation).toList();
  }

  bool _elementLacksDocumentation(ModelElement elem) {
    return !elem.hasDocumentation;
  }
}

enum EntityType {
  Class,
  Function,
  Variable,
  Enum
}

/// This class is returned if dartdoc fails in an expected way (for instance, if
/// there is an analysis error in the library).
class DartDocFailure {
  final String message;

  DartDocFailure(this.message);

  String toString() => message;
}

class _Error implements Comparable {
  final AnalysisError error;
  final LineInfo lineInfo;
  final String projectPath;

  _Error(this.error, this.lineInfo, this.projectPath);

  String get description => '${error.message} at ${location}, line ${line}.';
  bool get isError => error.errorCode.errorSeverity == ErrorSeverity.ERROR;
  int get line => lineInfo.getLocation(error.offset).lineNumber;
  String get location {
    String path = error.source.fullName;
    if (path.startsWith(projectPath)) {
      path = path.substring(projectPath.length + 1);
    }
    return path;
  }

  int get severity => error.errorCode.errorSeverity.ordinal;

  String get severityName => error.errorCode.errorSeverity.displayName;

  int compareTo(_Error other) {
    if (severity == other.severity) {
      int cmp = error.source.fullName.compareTo(other.error.source.fullName);
      return cmp == 0 ? line - other.line : cmp;
    } else {
      return other.severity - severity;
    }
  }

  String toString() => '[${severityName}] ${description}';
}
